// Copyright IBM Corp. 2015,2016. All Rights Reserved.
// Node module: wlpn-cli-health-analyzer
// US Government Users Restricted Rights - Use, duplication or disclosure
// restricted by GSA ADP Schedule Contract with IBM Corp.

'use strict';

var MemoryLeak = require('./memory-leak.js');
var MemoryUsage = require('./memory-usage.js');
var debug = require('./debug')('health-analyzer');
var express = require('express');
var jsonParser = require('body-parser').json({strict: false});
var mandatory = require('./util').mandatory;

module.exports = HealthAnalyzer;

/**
 * Messages POSTed to the health analyzer are forwarded to specific
 * analyzers, depending on the target type.
 *
 * Messages of type appmetrics:gc received from the runner are forwarded
 * to all the analyzers.
 *
 * Analyzers can send message to the controller using the
 * HealthAnalyzer#sendMessage method.
 *
 * XXX(sam) why isn't the post target in the URL path? That would be more RESTy,
 * and would allow this code to be mostly deleted, and for the two analyzers to
 * be first-class components themselves, listening on specific paths. I assume
 * this is for legacy reasons/compatibility with Java? Except, I think the REST
 * APIs here are recently implemented.
 */
function HealthAnalyzer(options) {
  this._options = options;
  this._console = options.logger || console;
  this._manager = mandatory(options.manager, 'manager not provided');
  this._listeners = {};
  this._eventListeners = {};

  // Server uses the host name as a part of key, must be lower case.
  this._hostNameId = mandatory(options.hostName.toLowerCase(),
    'missing hostNameId');

  this._adminPort = mandatory(options.adminPort, 'missing adminPort');
  this._serverPath = mandatory(options.serverPath, 'missing serverPath');
  this._serverName = mandatory(options.serverName, 'missing serverName');
  this._userDir = mandatory(options.userDir, 'missing userDir');
  debug('Host Name: %s, Admin Port: %s, Server Path: %s, Server Name: %s, '
        + 'User Dir: %s', this._hostNameId, this._adminPort, this._serverPath,
        this._serverName, this._userDir);
};

HealthAnalyzer.prototype.start = function(options, callback) {
  var self = this;

  debug('start');
  self._controller = self._manager.get('controller')();
  if (!self._controller || !self._controller.request) {
    return callback(new Error('no controller component was found'));
  }
  self._runner = self._manager.get('runner')();

  [MemoryUsage, MemoryLeak].forEach(function(Analyzer) {
    var a = new Analyzer(self._options, self);
    self._listeners[a.name] = a;
  });

  self._app = self._app || express(); // For unit testing.
  self._manager.app.use('/health', self._app);
  self._app.use(jsonParser);
  self._app.use(debug.json.middleware);

  self._app.post('/', function(req, res) {
    debug('recv POST %j', req.body);
    self._receiveMessage(req.body);
    res.send(); // Status code is ignored by client.
  });
  self._runner.on('request', function(msg) {
    debug('on request %s', msg.cmd);
    var events = Object.keys(self._eventListeners);
    for (var i = 0; i < events.length; i++) {
      if (events[i] === msg.cmd) {
        var listeners = self._eventListeners[events[i]];
        for (var j = 0; j < listeners.length; j++) {
          debug('updating listener: %s', listeners[j]);
          listeners[j](msg.data);
        }
      }
    }
  });
  self._establishConnectionWithParent(callback);
};

HealthAnalyzer.prototype.reconnect = function(server, cb) {
  this._establishConnectionWithParent(cb);
};

// Used by analyzers to report to the controller.
HealthAnalyzer.prototype.sendMessage = function(targetType, message, callback) {
  send(this, targetType, message, null, callback);
};

function send(self, targetType, message, messageType, callback) {
  var json = {
    source: [self._hostNameId, self._userDir, self._serverName].join(','),
    sourceAddress: self._hostNameId + ':' + self._adminPort,
  };
  if (message)
    json.message = message;
  // Case of targettype and messageType are inconsistent, but also used by the
  // Java member, do not change the case.
  // XXX(sam) confirm that they are really that inconsistent.
  if (targetType)
    json.targettype = targetType;
  if (messageType)
    json.messageType = messageType;

  debug('StatsCollector.sendMessage: %j', json);

  var data = {
    path: '/ibm/api/healthManager',
    verb: 'POST',
    data: json,
  };
  self._controller.request(data, function(err, res, data) {
    debug('Sending data to the controller: err= %s %s %j', err, res, data);
    return callback(err);
  });
};

HealthAnalyzer.prototype._receiveMessage = function(message) {
  debug('processing a message: %s', message);

   // targettype is also used by the Java member, do not change the case
  if (message.targettype && message.message) {
    var targetType = message.targettype;
    var msg = message.message;

    debug('processing a message - targettype = %s', targetType);
    debug('processing a message - msg = %s', msg);

    var listener = this._listeners[targetType];
    if (listener) {
      debug('Sending to plugin : %s', listener.name);
      listener.receiveMessage(msg);
    }
  }
};

HealthAnalyzer.prototype._establishConnectionWithParent = function(callback) {
  send(this, null, null, 'Initialization', callback);
};

HealthAnalyzer.prototype.addEventListener = function(cmd, listener) {
  debug('adding event listener for %s', cmd);
  var listeners = this._eventListeners[cmd];
  if (listeners) {
    listeners.push(listener);
  } else {
    listeners = [listener];
  }
  this._eventListeners[cmd] = listeners;
};
