// Copyright IBM Corp. 2015,2016. All Rights Reserved.
// Node module: wlpn-cli-health-analyzer
// US Government Users Restricted Rights - Use, duplication or disclosure
// restricted by GSA ADP Schedule Contract with IBM Corp.

'use strict';

var debug = require('./debug')('memory-leak');

module.exports = MemoryLeak;

function AverageMemData(heap, time) {
  this.heap = heap;
  this.time = time;
};

function MemoryLeak(options, collector) {
  this._console = options.logger || console;
  this._triggeredState = false;
  this._policyEnabled = false;
  this._sendMessage = collector.sendMessage.bind(collector);
  this._Gcs = [];
  this._RollingAvgsFreeHeap = [];
  this._RollingAvgsUsedHeap = [];
  this._Counter = 0;

  this.name = 'MemoryLeakServerCondition';

  this._minSamples = 100;
  this._compDistance = 50;
  this._initIgnore = 3;
  collector.addEventListener('appmetrics:gc', this.update.bind(this));
};

MemoryLeak.prototype._process = function() {
  debug('process');

  if (!this._policyEnabled)
    return;

  var recommendation = this._recommend();
  if (!recommendation) {
    this._triggeredState = false;
    return;
  }

  if (!this._triggeredState) {
    this._console.log('Memory leak threshold exceeded: memory leak %s',
                      recommendation);
    // reset counters to start new range for GC stats
    this._resetState('all');
    // notify handlers
    var message = {
      triggered: true,
      recommendation: recommendation,
      status: 1,
      condition: 'com.ibm.ws.health.manager.healthPolicy.condition.memoryLeak',
    };
    var self = this;
    // XXX(sam) why doesn't sendMessage do this serialization?
    var msg = JSON.stringify(message);
    this._sendMessage('ServerCondition', msg, function(err) {
      if (!err)
        self._triggeredState = true;
    });
  }
};

/**
 * Clears various Lists for new garbage collection stats - typically when
 * policy is violated and new set of data is required to calculate next
 * policy violation or a new set of rolling averages are needed
 *
 * @param type - used heap or free heap or all
 */
MemoryLeak.prototype._resetState = function(type) {
  switch (type) {
    case 'used':
      this._RollingAvgsUsedHeap.length = 0;
      break;
    case 'free':
      this._RollingAvgsFreeHeap.length = 0;
      break;
    case 'all':
    default:
      // reset counters to start new range for GC stats
      this._Gcs.length = 0;
      this._RollingAvgsFreeHeap.length = 0;
      this._RollingAvgsUsedHeap.length = 0;
      break;
  }
};

/**
 * Calculate rolling average of free and used heap.
 *
 * From Wikipedia: In statistics, a moving average (rolling average or running
 * average) is a calculation to analyze data points by creating a series of
 * averages of different subsets of the full data set.
 *
 * @param windowLength - length from which samples are taken
 */
MemoryLeak.prototype._calcRollingAverages = function(windowLength) {
  this._rollingAverageHelper(windowLength, 'used', this._RollingAvgsUsedHeap);
  this._rollingAverageHelper(windowLength, 'free', this._RollingAvgsFreeHeap);
};

MemoryLeak.prototype._rollingAverageHelper = _rollingAverageHelper;

function _rollingAverageHelper(windowLength, memType, rollingData) {
  var data = this._Gcs;

  if (data.length < windowLength) {
    return;
  }

  var finish = windowLength;
  var total = 0;
  var halfFinished = Math.floor(finish / 2);

  for (var i = 0; i < finish; i++) {
    var usedHeap = 0;
    if (memType === 'free') {
      usedHeap = data[i].size - data[i].used;
    } else {
      usedHeap = data[i].used;
    }
    total += usedHeap;
  }

  this._resetState(memType);

  rollingData.push(new AverageMemData(
    total / windowLength,
    data[halfFinished].time
  ));

  for (i = 0; i < data.length - windowLength; i++) {
    if (memType === 'free') {
      // Free heap is obtained by subtracting used heap from available heap.
      total -= (data[i].size - data[i].used);
      total += (data[finish + i].size - data[finish + i].used);
    } else {
      total -= data[i].used;
      total += data[finish + i].used;
    }
    var average = total / windowLength;
    var halfTime = data[halfFinished + i].time;
    rollingData.push(new AverageMemData(average, halfTime));
  }
};

MemoryLeak.prototype._recommend = function() {
  // XXX(sam) AFAICT, this is invariant:
  //   this._RollingAvgsFreeHeap.length === this._RollingAvgsUsedHeap.length
  // But the code below seems to think that they can be different.

  // XXX(sam) Node.js is known for using all memory up until its heap maximum
  // before starting to collect garbage, this algorithm below may need testing
  // against real systems to confirm its applicability.

  if (!this._RollingAvgsFreeHeap.length || !this._RollingAvgsUsedHeap.length)
    return;

  var sampleLength = Math.min(this._RollingAvgsFreeHeap.length,
                              this._RollingAvgsUsedHeap.length);

  debug('original RollingAvgsHeap size free %d used %d samepleLength %d',
        this._RollingAvgsFreeHeap.length,
        this._RollingAvgsUsedHeap.length,
        sampleLength);

  if (sampleLength < this._minSamples)
    return;

  var start = Math.floor(sampleLength / this._initIgnore);

  for (var i = start; i < sampleLength - this._compDistance; i++) {
    var amountFreedOne = this._RollingAvgsFreeHeap[i];
    var amountFreedTwo = this._RollingAvgsFreeHeap[i + this._compDistance];
    var usedHeapOne = this._RollingAvgsUsedHeap[i];
    var usedHeapTwo = this._RollingAvgsUsedHeap[i + this._compDistance];

    var changeInAmountFreed =
      (amountFreedOne.heap - amountFreedTwo.heap) /
      (amountFreedOne.time - amountFreedTwo.time);
    var changeInAmountUsed =
      (usedHeapOne.heap - usedHeapTwo.heap) /
      (usedHeapOne.time - usedHeapTwo.time);

    debug('amountFreedOne %j; amountFreedTwo %j',
          amountFreedOne.heap, amountFreedTwo.heap);
    debug('usedHeapOne %j; usedHeapTwo %j',
          usedHeapOne.heap, usedHeapTwo.heap);
    debug('change in amount freed = %j', changeInAmountFreed);
    debug('change in amount used = %j', changeInAmountUsed);

    if ((changeInAmountUsed <= 0) ||
        (changeInAmountUsed - changeInAmountFreed <= 0.2))
      continue;
    // else memory is leaking
    var memoryLeakMessage = 'Memory seems to be growing from ' +
      new Date(amountFreedOne.time) + ' to ' +
      new Date(amountFreedTwo.time) +
      ' with change in used amount = ' + changeInAmountUsed +
      ' and change in free memory = ' + changeInAmountFreed;
    return memoryLeakMessage;
  }
};

MemoryLeak.prototype.receiveMessage = function(msg) {
  var threshold = JSON.parse(msg);
  var enabled = threshold.operation;
  if (enabled != null) {
    if (enabled === 'set') {
      this._console.log('memory leak monitor is on');
      this._policyEnabled = true;
      this._triggeredState = false;
    } else {
      // anything else is false
      this._policyEnabled = false;
      this._resetState('all');
    }
  }
  this._console.log('Modified memory leak policy: enabled = %s',
                    this._policyEnabled);
};

// Receive appmetrics:gc message.data
// XXX(sam) form of data is...?
MemoryLeak.prototype.update = function(data) {
  // Start collecting stats only when policy is enabled.
  debug('memory leak: update %j', data);

  if (!this._policyEnabled) return;

  var gcType = data.type;
  if (gcType === 'S') { // XXX(sam) 'S' means what?
    this._Gcs.push(data);
    this._Counter++;
    this._calcRollingAverages(5); // XXX(sam) why is 5 not a property?
    this._process(data);
  }
  this._printSavedData();
};

MemoryLeak.prototype._printSavedData = function() {
  debug('regularRollingAvgsUsedHeap size %j', this._RollingAvgsUsedHeap.length);
  debug('regularRollingAvgsFreeHeap size %j', this._RollingAvgsFreeHeap.length);
  debug('regularCounter size %j', this._Counter);
};
