// Copyright IBM Corp. 2015,2016. All Rights Reserved.
// Node module: wlpn-cli-health-analyzer
// US Government Users Restricted Rights - Use, duplication or disclosure
// restricted by GSA ADP Schedule Contract with IBM Corp.

'use strict';

var debug = require('./debug')('memory-usage');

module.exports = MemoryUsage;

function MemoryUsage(options, collector) {
  this._console = options.logger || console;
  this._triggeredState = false;
  this.name = 'MemoryUsageServerCondition';
  this._policyEnabled = false;
  this._sendMessage = collector.sendMessage.bind(collector);
  this._firstThresCrossing = undefined;
  this._lastUpdateTime = 0;

  this._memThresPercent = 0;
  this._maxHeapSizeMB = 0;
  this._maxTime = 0;
  collector.addEventListener('appmetrics:gc', this.update.bind(this));
};

MemoryUsage.prototype.process = function(memoryUsage) {
  debug('process');

  if (!this._policyEnabled)
    return;

  if (!this._thresholdExceeded(memoryUsage))
    return;

  this._console.log(
    'Memory usage threshold exceeded: memory usage %j, used heap %j',
    memoryUsage.size, memoryUsage.used);

  var message = {
    triggered: true,
    heap: memoryUsage.size / 1e+6, // Converting from bytes to megabytes
    usedheap: memoryUsage.used / 1e+6,
    status: 1,
    condition: 'com.ibm.ws.health.manager.healthPolicy.condition.memoryUsage',
  };

  var self = this;
  var msg = JSON.stringify(message);
  this._sendMessage('ServerCondition', msg, function(err) {
    if (!err)
      self._firstThresCrossing = undefined;
  });
};

MemoryUsage.prototype._thresholdExceeded = function(stats) {
  var curtime = Date.now();
  var usedHeapPercentage = (stats.used / stats.size) * 100;
  this._lastUpdateTime = curtime;
  var triggered = false;

  debug('Defined memory usage threshold %j', this._memThresPercent);
  debug('Defined memory usage duration %j', this._maxTime);
  debug('used heap percentage %j', usedHeapPercentage);

  debug('Current time %j', curtime);
  debug('firstThresCrossing %j', this._firstThresCrossing);
  debug('elapsed time since first threshold crossing (ms) %j',
        curtime - this._firstThresCrossing);

  if (usedHeapPercentage >= this._memThresPercent && stats.size) {
    if (!this._triggeredState) {
      if (this._firstThresCrossing &&
          (curtime - this._firstThresCrossing > this._maxTime)) {
        this._triggeredState = true;
        triggered = true;
      } else if (this._firstThresCrossing === undefined) {
        this._firstThresCrossing = curtime;
        // Start timer in case we do not get any GC events.
        this._startTimer(stats, this._maxTime);
      } else {
        // Start timer for the remaining duration.
        //
        // This is needed for cases when GC is not triggered following a heap
        // violation, it is entirely possible that GC stabilized after exceeding
        // our threshold so we would not get any further GC events meaning we
        // still need to track the duration (maxTime) that was specified in the
        // policy and trigger the policy.
        this._startTimer(stats, this._maxTime -
                         (curtime - this._firstThresCrossing));
      }
    }
  } else {
    // Below the threshold, clear and start over.
    this._firstThresCrossing = undefined;
    this._cancelTimer();
    this._triggeredState = false;
  }

  debug('Mem (MB)='
    + stats.used / 1e6 + ' Available (MB)=' + stats.size - stats.use / 1e6 +
    ' Max Mem (MB)=' + stats.size / 1e6 +
    ' Percent Used Mem=' + usedHeapPercentage
    + ' Policy percent threshold is=' + this._memThresPercent
    + ' curtime=' + Date(curtime) + ' firstThresCrossing=' +
    ((this._firstThresCrossing === undefined) ? 'undefined' :
    Date(this._firstThresCrossing)) + ' triggered=' + triggered);

  return triggered;
};

MemoryUsage.prototype.receiveMessage = function(msg) {
  try {
    var threshold = JSON.parse(msg);
  } catch (e) {
    // Invalid message, ignore it.
    debug('Failed to process received message: %s', e.message);
    return;
  }

  debug('threshold: %j', threshold);

  var tp = threshold.thresholdPercent;
  var mt = threshold.maxTime;
  this._memThresPercent = tp != null ? tp : 100;
  this._maxTime = mt != null ? mt : 0;

  this._console.log('Modified memory usage threshold %j',
                  this._memThresPercent);
  this._console.log('Modified memory usage duration %j', this._maxTime);
  debug('Maximum Heap %j', this._maxHeapSizeMB);

  this._policyEnabled = true;
  this._triggeredState = false;
};

MemoryUsage.prototype._startTimer = function(memoryUsage, interval) {
  var process = this.process.bind(this, memoryUsage);
  this._scheduledFuture = setTimeout(process, interval * 1000).unref();
  debug('setTimer: %j', interval * 1000);
};

MemoryUsage.prototype._cancelTimer = function() {
  clearTimeout(this._scheduledFuture);
  debug('cancelTimer');
};

MemoryUsage.prototype.update = function(data) {
  // Start collecting stats only when policy is enabled.
  if (!this._policyEnabled)
    return;

  var gcType = data.type;
  if (gcType === 'S') {
    this._cancelTimer(); // Cancel timer if running, we want to start fresh.
    this.process(data);
  }
};
